﻿/* 
 * Copyright (C) 2008 Alan Ross, Tojio GmbH (http://lab.tojio.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
*/
 

package
{	
	import flash.display.*;
	import flash.events.*;
	import flash.text.TextField;
	//import Papervision (Great White)
	import org.papervision3d.cameras.Camera3D;
	import org.papervision3d.materials.BitmapMaterial;
	import org.papervision3d.materials.WireframeMaterial;
	import org.papervision3d.materials.utils.MaterialsList;
	import org.papervision3d.objects.DisplayObject3D;
	import org.papervision3d.objects.primitives.Cube;
	import org.papervision3d.view.BasicView;
	//import WiiFlash (0.4)
	import org.wiiflash.*;
	import org.wiiflash.events.*;

    //[SWF(frameRate="31", width="900", height="600", backgroundColor="0x111111")]
    public class WiiFlash3DObjectNunchuck extends Sprite
    {
    	//papervision
        private var view:BasicView;
        private var rootNode:DisplayObject3D;
        private var environment:Cube;
        
        //Play around with this value to get rid of the 'noise' the Wiimote sends
        //I guess each Wiimote will have its own optimal value here... 
		private const THRESHOLD:Number = .0003;
		private const ROTATION_FACTOR:Number = 5;
		private const MOVE_FACTOR:Number = 2;
		
		private var newRotX:Number=0;
		private var oldRotX:Number=0;
		private var newRotY:Number=0;
		private var oldRotY:Number=0;
		private var newRotZ:Number=0;
		private var oldRotZ:Number=0;
		
		private var newMoveX:Number=0;
		private var oldMoveX:Number=0;
		private var newMoveY:Number=0;
		private var oldMoveY:Number=0;
		
		// create a new Wiimote
		private var wiimote:Wiimote;
		private var hasNunchuck:Boolean;
        
		//debugging
		private var debug:TextField;
		
		//________________________________________________________
		// constructor
		//________________________________________________________
		public function WiiFlash3DObjectNunchuck()
		{   
            super();
			if(stage != null){ init(); }
			else{ addEventListener(Event.ADDED_TO_STAGE, init); }
		}
		
		
		//________________________________________________________
		// initialize
		//________________________________________________________
		public function init(e:Event=null):void 
		{	
			stage.scaleMode = StageScaleMode.NO_SCALE;
        	stage.align = StageAlign.TOP_LEFT;
        	stage.quality = StageQuality.HIGH;
        	
        	debug = new TextField(); debug.height = 400; debug.width = 300; 
        	debug.x = debug.y = 10; debug.textColor = 0xffffff; debug.selectable = false; 
        	addChild(debug);
        	
        	initWiiFlash();
        	init3D();
        	
        	addEventListener( Event.ENTER_FRAME, loop );
		}
		
		
		//________________________________________________________
		// setup papervision
		//________________________________________________________
		private function init3D():void
		{
			view = new BasicView(stage.stageWidth, stage.stageHeight, true, true);
        	view.camera.zoom = 12;
        	view.camera.focus = 100;
        	addChildAt(view,0);
	
			//the root node
			rootNode = new DisplayObject3D();
			
			//the environment
			var m:WireframeMaterial = new WireframeMaterial(0x333333,.2);
			var ml:MaterialsList = new MaterialsList({front:m, back:m, right:m, left:m, top:m, bottom:m});
			var s:uint = 700;	//size
			var q:uint = 10;	//quality
			environment = new Cube(ml,s,s,s,q,q,q,Cube.ALL);
			view.scene.addChild(environment, "environment");
			
			//now add a demo object the Wiimote can play with...
			var bmd:BitmapData = new BitmapData(1000,1000);
			bmd.perlinNoise(10,10,3,Math.random()*10,true,true);
			var bm:BitmapMaterial = new BitmapMaterial(bmd);
			
			rootNode.addChild(new Cube(new MaterialsList({all:bm}),100,10,100,10,1,10));
			
        	view.scene.addChild(rootNode);
		}
		
		//________________________________________________________
		// setup papervision
		//________________________________________________________
		private function initWiiFlash():void
		{
			hasNunchuck = false;
			wiimote = new Wiimote();
			wiimote.connect();
			
			wiimote.addEventListener( Event.CONNECT, onWiimoteConnect );
			wiimote.addEventListener( IOErrorEvent.IO_ERROR, onWiimoteConnectError );
			wiimote.addEventListener( Event.CLOSE, onCloseConnection );
			wiimote.addEventListener( WiimoteEvent.UPDATE, onUpdated );
			wiimote.addEventListener( WiimoteEvent.NUNCHUK_CONNECT, onNunchukConnected );
			wiimote.addEventListener( WiimoteEvent.NUNCHUK_DISCONNECT, onNunchukDisconnected );
		}
		
		//________________________________________________________
		// main rendering loop
		//________________________________________________________
        private function loop( event : Event ):void
        {
        	//Use mouse to fly around a little
        	//Camera3D(view.camera).hover(0, view.viewport.containerSprite.mouseX*-.004, view.viewport.containerSprite.mouseY*.004);
        	//update rotation and pos of 3d object
			update3DObject();
			//render all changes done
			view.singleRender();
        }

		
		//________________________________________________________
		// update rotation and pos of 3d object
		//________________________________________________________
       	private function update3DObject():void
		{
			if(hasNunchuck)
			{
				newRotX = wiimote.nunchuk.stickY
        		newRotY = wiimote.nunchuk.stickX;
        		newRotZ = 0;
				newRotX = (newRotX+oldRotX)/2;
				newRotY = (newRotY+oldRotY)/2;
				newRotZ = (newRotZ+oldRotZ)/2;
				
				newMoveX = wiimote.nunchuk.sensorX;
				newMoveY = wiimote.nunchuk.sensorY;
				newMoveX = (newMoveX+oldMoveX)/2;
				newMoveY = (newMoveY+oldMoveY)/2;
			}
        	
			rootNode.rotationX -=(180/Math.PI*(newRotX/ROTATION_FACTOR));
        	rootNode.rotationY -=(180/Math.PI*(newRotY/ROTATION_FACTOR));
        	rootNode.rotationZ -=(180/Math.PI*(newRotZ/ROTATION_FACTOR));
        	
			if(Math.abs(newMoveX-oldMoveX) > THRESHOLD && Math.abs(newMoveY-oldMoveY) > THRESHOLD)
			{
				Camera3D(view.camera).hover(0, newMoveX*-MOVE_FACTOR, newMoveY*-MOVE_FACTOR);
			}
			
			if(wiimote.nunchuk.c) if(rootNode.scale <5) rootNode.scale += .1;
			if(wiimote.nunchuk.z) if(rootNode.scale >1) rootNode.scale -= .1;
			
			oldRotX = newRotX;
			oldRotY = newRotY;
			oldRotZ = newRotZ;
			oldMoveX = newMoveX;
			oldMoveY = newMoveY;
		}

		// just some info from here on...
		private function onCloseConnection( e:Event ):void
		{ debug.text = "Connection closed, WiiFlash Server has been closed"; }
		
		private function onWiimoteConnectError( e:IOErrorEvent ):void
		{ debug.text = "Couldn't connect, make sure WiiFlash Server is running !"; }
	
		private function onWiimoteConnect( e:Event ):void
		{ debug.text= "Wiimote successfully connected"; }
	
		private function onUpdated( e:WiimoteEvent ):void 
		{
			debug.text = "Wiimote: "+wiimote.id+"\n";
			debug.appendText("Nunchuk connected: " + hasNunchuck + "\n");
			debug.appendText("Nunchuk Sensor X: " + wiimote.nunchuk.sensorX + "\n");
			debug.appendText("Nunchuk Sensor Y: " + wiimote.nunchuk.sensorY + "\n");
			debug.appendText("Nunchuk Stick X: " + wiimote.nunchuk.stickX + "\n");
			debug.appendText("Nunchuk Stick Y: " + wiimote.nunchuk.stickY + "\n");
			
			debug.appendText("Nunchuck Button C: " + wiimote.nunchuk.c + "\n");
			debug.appendText("Nunchuck Button Z: " + wiimote.nunchuk.z + "\n");
		}
		
		private function onNunchukConnected( e:Event ):void
		{ debug.text= "Nunchuck is connected"; hasNunchuck = true; }
		
		private function onNunchukDisconnected( e:Event ):void
		{ debug.text= "Nunchuck is disconnected"; hasNunchuck = false; }
    }
}